﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;

namespace RumbleComponentExample
{
    struct RumbleInstance
    {
        //Control over the motors and for how long.
        public float LeftMotorAmount;
        public float RightMotorAmount;
        public float TimeLeft;

        public PlayerIndex playerIndex;

        //This instance's state members. These define if it is alive, 
        //for how long, and the shape of the rumble.
        public bool IsAlive { get; private set; }
        public Rumble.Shape Shape { get; private set; }
        public float Time { get; private set; }

        //This initializes this instance with these attributes.
        public void Rumble(
            PlayerIndex playerIndex,
            float leftMotorAmount, 
            float rightMotorAmount, 
            float rumbleTime,
            Rumble.Shape Shape)
        {
            if (rumbleTime <= 0)
                throw new Exception("rumbleTime must be greater than zero");

            IsAlive = true;
            LeftMotorAmount = MathHelper.Clamp(leftMotorAmount, 0 ,1);
            RightMotorAmount = MathHelper.Clamp(rightMotorAmount, 0, 1);
            TimeLeft = rumbleTime;
            Time = rumbleTime;

            this.playerIndex = playerIndex;
            this.Shape = Shape;
        }

        //Set this instance to null values.
        public void Kill()
        {
            IsAlive = false;
            LeftMotorAmount = 0;
            RightMotorAmount = 0;
            TimeLeft = 0;
            Time = 0;
        }
    }

    public class Rumble : GameComponent
    {
        public enum Shape
        {
            Binary,             //Motors On/Off at max
            Linear,             //Start at zero, Linear increase, End at max.
            LinearInvert,       //Start at max, Linear decrease, End at zero.
            Exponential,        //Start at zero, Exponential increase, End at max.
            ExponentialInvert,  //Start at max, Exponential decrease, End at zero.
            SmoothStep,         //Start at zero, Sigmoidal increase, End at max.
            SmoothStepInvert,   //Start at max, Sigmoidal decrease, End at zero.
            Parabolic           //Motors gradually On/Off - Apex at time / 2.
        }

        //Make a pool of rumble instances.
        RumbleInstance[] rumblePool = { 
                                         new RumbleInstance(), 
                                         new RumbleInstance(),
                                         new RumbleInstance(),
                                         new RumbleInstance(),
                                         new RumbleInstance(),
                                         new RumbleInstance(),
                                         new RumbleInstance(),
                                         new RumbleInstance() };

        public Rumble(Game game)
            : base(game)
        {
            for (int i = 0; i < rumblePool.Length; i++)
                rumblePool[i].Kill();
        }

        public override void Update(GameTime gameTime)
        {
            for (int i = 0; i < rumblePool.Length; i++)
            {
                if (rumblePool[i].IsAlive)
                {
                    //Reduce the time left to vibrate the motors, 
                    //but don't go less than zero.
                    rumblePool[i].TimeLeft =
                        MathHelper.Clamp(
                            rumblePool[i].TimeLeft - 
                            (float)gameTime.ElapsedGameTime.TotalSeconds, 
                            0, 
                            rumblePool[i].Time);

                    //Reduce or shape the "curve" of the vibration.
                    Vector2 motorSpeeds = ShapeRumble(rumblePool[i]);

                    if (rumblePool[i].TimeLeft <= 0)
                    {
                        rumblePool[i].Kill();           //Reset the rumble properties. 
                        motorSpeeds = Vector2.Zero;     //Stop the active motors.
                    }

                    GamePad.SetVibration(
                        rumblePool[i].playerIndex, motorSpeeds.X, motorSpeeds.Y);
                }
            }

            base.Update(gameTime);
        }

        private Vector2 ShapeRumble(RumbleInstance rumble)
        {
            Vector2 motorSpeeds = Vector2.Zero;

            //We're going to use rumble.Time as a divisor so error check
            //this in case it ever gets less than or equal to zero.
            //It should never reach this code if in that senerio, but this is just a failsafe.
            if (rumble.Time <= 0)
                return motorSpeeds;

            switch (rumble.Shape)
            {
                case Shape.Binary:
                    motorSpeeds = Binary(rumble);
                    break;
                case Shape.Linear:
                    motorSpeeds = Linear(rumble);
                    break;
                case Shape.LinearInvert:
                    motorSpeeds = LinearInvert(rumble);
                    break;
                case Shape.Exponential:
                    motorSpeeds = Exponential(rumble);
                    break;
                case Shape.ExponentialInvert:
                    motorSpeeds = ExponentialInvert(rumble);
                    break;
                case Shape.SmoothStep:
                    motorSpeeds = SmoothStep(rumble);
                    break;
                case Shape.SmoothStepInvert:
                    motorSpeeds = SmoothStepInvert(rumble);
                    break;
                case Shape.Parabolic:
                    motorSpeeds = Parabolic(rumble);
                    break;
            }

            return motorSpeeds;
        }

        /// <summary>
        /// Add a rumble with the default shape.
        /// </summary>
        /// <param name="playerIndex">The player controller to rumble.</param>
        /// <param name="leftMotor">The maximum value of the left motor, between 0.0f and 1.0f.</param>
        /// <param name="rightMotor">The maximum value of the right motor, between 0.0f and 1.0f.</param>
        /// <param name="time">The amount of time to rumble the motor in seconds.</param>
        public void AddRumble(PlayerIndex playerIndex, float leftMotor, float rightMotor, float time)
        {
            AddRumble(playerIndex, leftMotor, rightMotor, time, Shape.Binary);
        }

        /// <summary>
        /// Add a rumble with a predefined shape.
        /// </summary>
        /// <param name="playerIndex">The player controller to rumble.</param>
        /// <param name="leftMotor">The maximum value of the 
        /// left motor, between 0.0f and 1.0f.</param>
        /// <param name="rightMotor">The maximum value of the right motor, between 0.0f and 1.0f.</param>
        /// <param name="time">The amount of time to rumble the motor in seconds.</param>
        /// <param name="shape">The shape of the rumble.</param>
        public void AddRumble(PlayerIndex playerIndex, float leftMotor, float rightMotor, float time, Shape shape)
        {
            //Find the first unused rumble instance and use it.
            //If one is not available, rumble will not be set.
            for (int i = 0; i < rumblePool.Length; i++)
            {
                if (!rumblePool[i].IsAlive)
                {
                    //Use this instance of a rumble and break.
                    rumblePool[i].Rumble(playerIndex, leftMotor, rightMotor, time, shape);
                    break;
                }
            }
        }

        private Vector2 Binary(RumbleInstance rumble)
        {
            //Rumble full blast until out of time, so leave the motors unchanged.
            return new Vector2(rumble.LeftMotorAmount, rumble.RightMotorAmount);
        }

        private Vector2 Linear(RumbleInstance rumble)
        {
            //Start rumble at zero and linearly increase force feedback 
            //over the length of time until motorAmount is reached.

            Vector2 motorSpeeds = Vector2.Zero;

            motorSpeeds.X = MathHelper.Lerp(0, rumble.LeftMotorAmount, 1 - (rumble.TimeLeft / rumble.Time));
            motorSpeeds.Y = MathHelper.Lerp(0, rumble.RightMotorAmount, 1 - (rumble.TimeLeft / rumble.Time));

            return motorSpeeds;
        }

        private Vector2 LinearInvert(RumbleInstance rumble)
        {
            //Start rumble at motorAmount and linearly decrease force feedback 
            //over the length of time until zero is reached.

            Vector2 motorSpeeds = Vector2.Zero;

            motorSpeeds.X = MathHelper.Lerp(0, rumble.LeftMotorAmount, rumble.TimeLeft / rumble.Time);
            motorSpeeds.Y = MathHelper.Lerp(0, rumble.RightMotorAmount, rumble.TimeLeft / rumble.Time);

            return motorSpeeds;
        }

        private Vector2 Exponential(RumbleInstance rumble)
        {
            //Start rumble at zero and exponentially increase force feedback 
            //over the length of time until motorAmount is reached.

            Vector2 motorSpeeds = Vector2.Zero;

            float amount = (float)Math.Pow(1 - (rumble.TimeLeft / rumble.Time), 2);

            motorSpeeds.X = MathHelper.Lerp(0, rumble.LeftMotorAmount, amount);
            motorSpeeds.Y = MathHelper.Lerp(0, rumble.RightMotorAmount, amount);

            return motorSpeeds;
        }

        private Vector2 ExponentialInvert(RumbleInstance rumble)
        {
            //Start rumble at motorAmount and exponentially decrease force feedback 
            //over the length of time until zero is reached.

            Vector2 motorSpeeds = Vector2.Zero;

            float amount = (float)Math.Pow(rumble.TimeLeft / rumble.Time, 2);

            motorSpeeds.X = MathHelper.Lerp(0, rumble.LeftMotorAmount, amount);
            motorSpeeds.Y = MathHelper.Lerp(0, rumble.RightMotorAmount, amount);

            return motorSpeeds;
        }

        private Vector2 SmoothStep(RumbleInstance rumble)
        {
            //Start rumble at zero and sigmoidally increase force feedback 
            //over the length of time until motorAmount is reached.

            Vector2 motorSpeeds = Vector2.Zero;

            motorSpeeds.X = MathHelper.SmoothStep(0, rumble.LeftMotorAmount, 1 - (rumble.TimeLeft / rumble.Time));
            motorSpeeds.Y = MathHelper.SmoothStep(0, rumble.RightMotorAmount, 1 - (rumble.TimeLeft / rumble.Time));

            return motorSpeeds;
        }

        private Vector2 SmoothStepInvert(RumbleInstance rumble)
        {
            //Start rumble at motorAmount and sigmoidally decrease force feedback 
            //over the length of time until zero is reached.

            Vector2 motorSpeeds = Vector2.Zero;

            motorSpeeds.X = MathHelper.SmoothStep(0, rumble.LeftMotorAmount, rumble.TimeLeft / rumble.Time);
            motorSpeeds.Y = MathHelper.SmoothStep(0, rumble.RightMotorAmount, rumble.TimeLeft / rumble.Time);

            return motorSpeeds;
        }

        private Vector2 Parabolic(RumbleInstance rumble)
        {
            //Start rumble at zero and gradually increase force feedback.
            //When rumble.Time / 2.0 is reached motors will be at motorAmount. 
            //Then, gradually reduce force feedback back to zero.

            Vector2 motorSpeeds = Vector2.Zero;

            //Use the parabolic equation: 4x(1-x)
            //This gives us a curve that starts at zero when x = 0, 
            //reaches its apex at x = 0.5, and ends at zero when x = 1.
            float amount = 1 - (rumble.TimeLeft / rumble.Time);
            amount = 4 * amount * (1 - amount);

            motorSpeeds.X = MathHelper.Lerp(0, rumble.LeftMotorAmount, amount);
            motorSpeeds.Y = MathHelper.Lerp(0, rumble.RightMotorAmount, amount);

            return motorSpeeds;
        }
    }
}

